@@ -1,5 +1,3 @@
-from __future__ import print_function
-
 from collections import Counter, defaultdict, deque
 from functools import partial, wraps
 from heapq import merge
@@ -12,18 +10,14 @@ from itertools import (
     groupby,
     islice,
     repeat,
+    starmap,
     takewhile,
-    tee
+    tee,
+    zip_longest,
 )
 from operator import itemgetter, lt, gt, sub
 from sys import maxsize, version_info
-try:
-    from collections.abc import Sequence
-except ImportError:
-    from collections import Sequence
-
-from six import binary_type, string_types, text_type
-from six.moves import filter, map, range, zip, zip_longest
+from collections.abc import Sequence
 
 from .recipes import consume, flatten, take
 
@@ -52,6 +46,7 @@ __all__ = [
     'intersperse',
     'islice_extended',
     'iterate',
+    'last',
     'locate',
     'lstrip',
     'make_decorator',
@@ -60,6 +55,8 @@ __all__ = [
     'one',
     'padded',
     'peekable',
+    'replace',
+    'rlocate',
     'rstrip',
     'run_length',
     'seekable',
@@ -70,10 +67,13 @@ __all__ = [
     'split_at',
     'split_after',
     'split_before',
+    'split_into',
     'spy',
     'stagger',
     'strip',
+    'substrings',
     'unique_to_each',
+    'unzip',
     'windowed',
     'with_iter',
     'zip_offset',
@@ -136,7 +136,33 @@ def first(iterable, default=_marker):
         return default
 
 
-class peekable(object):
+def last(iterable, default=_marker):
+    """Return the last item of *iterable*, or *default* if *iterable* is
+    empty.
+
+        >>> last([0, 1, 2, 3])
+        3
+        >>> last([], 'some default')
+        'some default'
+
+    If *default* is not provided and there are no items in the iterable,
+    raise ``ValueError``.
+    """
+    try:
+        try:
+            # Try to access the last item directly
+            return iterable[-1]
+        except (TypeError, AttributeError, KeyError):
+            # If not slice-able, iterate entirely using length-1 deque
+            return deque(iterable, maxlen=1)[0]
+    except IndexError:  # If the iterable was empty
+        if default is _marker:
+            raise ValueError('last() was called on an empty iterable, and no '
+                             'default value was provided.')
+        return default
+
+
+class peekable:
     """Wrap an iterator to allow lookahead and prepending elements.
 
     Call :meth:`peek` on the result to get the value that will be returned
@@ -208,10 +234,6 @@ class peekable(object):
             return False
         return True
 
-    def __nonzero__(self):
-        # For Python 2 compatibility
-        return self.__bool__()
-
     def peek(self, default=_marker):
         """Return the item that will be next returned from ``next()``.
 
@@ -265,8 +287,6 @@ class peekable(object):
 
         return next(self._it)
 
-    next = __next__  # For Python 2 compatibility
-
     def _get_slice(self, index):
         # Normalize the slice's arguments
         step = 1 if (index.step is None) else index.step
@@ -306,14 +326,11 @@ class peekable(object):
         return self._cache[index]
 
 
-def _collate(*iterables, **kwargs):
+def _collate(*iterables, key=lambda a: a, reverse=False):
     """Helper for ``collate()``, called when the user is using the ``reverse``
     or ``key`` keyword arguments on Python versions below 3.5.
 
     """
-    key = kwargs.pop('key', lambda a: a)
-    reverse = kwargs.pop('reverse', False)
-
     min_or_max = partial(max if reverse else min, key=itemgetter(0))
     peekables = [peekable(it) for it in iterables]
     peekables = [p for p in peekables if p]  # Kill empties.
@@ -351,9 +368,7 @@ def collate(*iterables, **kwargs):
     If the elements of the passed-in iterables are out of order, you might get
     unexpected results.
 
-    On Python 2.7, this function delegates to :func:`heapq.merge` if neither
-    of the keyword arguments are specified. On Python 3.5+, this function
-    is an alias for :func:`heapq.merge`.
+    On Python 3.5+, this function is an alias for :func:`heapq.merge`.
 
     """
     if not kwargs:
@@ -409,12 +424,12 @@ def ilen(iterable):
     This consumes the iterable, so handle with care.
 
     """
-    # maxlen=1 only stores the last item in the deque
-    d = deque(enumerate(iterable, 1), maxlen=1)
-    # since we started enumerate at 1,
-    # the first item of the last pair will be the length of the iterable
-    # (assuming there were items)
-    return d[0][0] if d else 0
+    # This approach was selected because benchmarks showed it's likely the
+    # fastest of the known implementations at the time of writing.
+    # See GitHub tracker: #236, #230.
+    counter = count()
+    deque(zip(iterable, counter), maxlen=0)
+    return next(counter)
 
 
 def iterate(func, start):
@@ -442,8 +457,7 @@ def with_iter(context_manager):
 
     """
     with context_manager as iterable:
-        for item in iterable:
-            yield item
+        yield from iterable
 
 
 def one(iterable, too_short=None, too_long=None):
@@ -524,33 +538,28 @@ def distinct_permutations(iterable):
     sequence.
 
     """
-    def perm_unique_helper(item_counts, perm, i):
-        """Internal helper function
+    def make_new_permutations(permutations, e):
+        """Internal helper function.
+        The output permutations are built up by adding element *e* to the
+        current *permutations* at every possible position.
+        The key idea is to keep repeated elements (reverse) ordered:
+        if e1 == e2 and e1 is before e2 in the iterable, then all permutations
+        with e1 before e2 are ignored.
 
-        :arg item_counts: Stores the unique items in ``iterable`` and how many
-            times they are repeated
-        :arg perm: The permutation that is being built for output
-        :arg i: The index of the permutation being modified
-
-        The output permutations are built up recursively; the distinct items
-        are placed until their repetitions are exhausted.
         """
-        if i < 0:
-            yield tuple(perm)
-        else:
-            for item in item_counts:
-                if item_counts[item] <= 0:
-                    continue
-                perm[i] = item
-                item_counts[item] -= 1
-                for x in perm_unique_helper(item_counts, perm, i - 1):
-                    yield x
-                item_counts[item] += 1
+        for permutation in permutations:
+            for j in range(len(permutation)):
+                yield permutation[:j] + [e] + permutation[j:]
+                if permutation[j] == e:
+                    break
+            else:
+                yield permutation + [e]
 
-    item_counts = Counter(iterable)
-    length = sum(item_counts.values())
+    permutations = [[]]
+    for e in iterable:
+        permutations = make_new_permutations(permutations, e)
 
-    return perm_unique_helper(item_counts, [None] * length, length - 1)
+    return (tuple(t) for t in permutations)
 
 
 def intersperse(e, iterable, n=1):
@@ -630,6 +639,15 @@ def windowed(seq, n, fillvalue=None, step=1):
         >>> list(windowed([1, 2, 3, 4, 5, 6], 3, fillvalue='!', step=2))
         [(1, 2, 3), (3, 4, 5), (5, 6, '!')]
 
+    To slide into the iterable's items, use :func:`chain` to add filler items
+    to the left:
+
+        >>> iterable = [1, 2, 3, 4]
+        >>> n = 3
+        >>> padding = [None] * (n - 1)
+        >>> list(windowed(chain(padding, iterable), 3))
+        [(None, None, 1), (None, 1, 2), (1, 2, 3), (2, 3, 4)]
+
     """
     if n < 0:
         raise ValueError('n must be >= 0')
@@ -664,7 +682,33 @@ def windowed(seq, n, fillvalue=None, step=1):
         yield tuple(window)
 
 
-class bucket(object):
+def substrings(iterable):
+    """Yield all of the substrings of *iterable*.
+
+        >>> [''.join(s) for s in substrings('more')]
+        ['m', 'o', 'r', 'e', 'mo', 'or', 're', 'mor', 'ore', 'more']
+
+    Note that non-string iterables can also be subdivided.
+
+        >>> list(substrings([0, 1, 2]))
+        [(0,), (1,), (2,), (0, 1), (1, 2), (0, 1, 2)]
+
+    """
+    # The length-1 substrings
+    seq = []
+    for item in iter(iterable):
+        seq.append(item)
+        yield (item,)
+    seq = tuple(seq)
+    item_count = len(seq)
+
+    # And the rest
+    for n in range(2, item_count + 1):
+        for i in range(item_count - n + 1):
+            yield seq[i:i + n]
+
+
+class bucket:
     """Wrap *iterable* and return an object that buckets it iterable into
     child iterables based on a *key* function.
 
@@ -849,7 +893,7 @@ def collapse(iterable, base_type=None, levels=None):
     def walk(node, level):
         if (
             ((levels is not None) and (level > levels)) or
-            isinstance(node, string_types) or
+            isinstance(node, str) or
             ((base_type is not None) and isinstance(node, base_type))
         ):
             yield node
@@ -865,8 +909,7 @@ def collapse(iterable, base_type=None, levels=None):
                 for x in walk(child, level + 1):
                     yield x
 
-    for x in walk(iterable, 0):
-        yield x
+    yield from walk(iterable, 0)
 
 
 def side_effect(func, iterable, chunk_size=None, before=None, after=None):
@@ -924,8 +967,7 @@ def side_effect(func, iterable, chunk_size=None, before=None, after=None):
         else:
             for chunk in chunked(iterable, chunk_size):
                 func(chunk)
-                for item in chunk:
-                    yield item
+                yield from chunk
     finally:
         if after is not None:
             after()
@@ -1012,6 +1054,51 @@ def split_after(iterable, pred):
         yield buf
 
 
+def split_into(iterable, sizes):
+    """Yield a list of sequential items from *iterable* of length 'n' for each
+    integer 'n' in *sizes*.
+
+        >>> list(split_into([1,2,3,4,5,6], [1,2,3]))
+        [[1], [2, 3], [4, 5, 6]]
+
+    If the sum of *sizes* is smaller than the length of *iterable*, then the
+    remaining items of *iterable* will not be returned.
+
+        >>> list(split_into([1,2,3,4,5,6], [2,3]))
+        [[1, 2], [3, 4, 5]]
+
+    If the sum of *sizes* is larger than the length of *iterable*, fewer items
+    will be returned in the iteration that overruns *iterable* and further
+    lists will be empty:
+
+        >>> list(split_into([1,2,3,4], [1,2,3,4]))
+        [[1], [2, 3], [4], []]
+
+    When a ``None`` object is encountered in *sizes*, the returned list will
+    contain items up to the end of *iterable* the same way that itertools.slice
+    does:
+
+        >>> list(split_into([1,2,3,4,5,6,7,8,9,0], [2,3,None]))
+        [[1, 2], [3, 4, 5], [6, 7, 8, 9, 0]]
+
+    :func:`split_into` can be useful for grouping a series of items where the
+    sizes of the groups are not uniform. An example would be where in a row
+    from a table, multiple columns represent elements of the same feature
+    (e.g. a point represented by x,y,z) but, the format is not the same for
+    all columns.
+    """
+    # convert the iterable argument into an iterator so its contents can
+    # be consumed by islice in case it is a generator
+    it = iter(iterable)
+
+    for size in sizes:
+        if size is None:
+            yield list(it)
+            return
+        else:
+            yield list(islice(it, size))
+
+
 def padded(iterable, fillvalue=None, n=None, next_multiple=False):
     """Yield the elements from *iterable*, followed by *fillvalue*, such that
     at least *n* items are emitted.
@@ -1030,8 +1117,7 @@ def padded(iterable, fillvalue=None, n=None, next_multiple=False):
     """
     it = iter(iterable)
     if n is None:
-        for item in chain(it, repeat(fillvalue)):
-            yield item
+        yield from chain(it, repeat(fillvalue))
     elif n < 1:
         raise ValueError('n must be at least 1')
     else:
@@ -1108,7 +1194,7 @@ def stagger(iterable, offsets=(-1, 0, 1), longest=False, fillvalue=None):
     )
 
 
-def zip_offset(*iterables, **kwargs):
+def zip_offset(*iterables, offsets, longest=False, fillvalue=None):
     """``zip`` the input *iterables* together, but offset the `i`-th iterable
     by the `i`-th item in *offsets*.
 
@@ -1116,7 +1202,7 @@ def zip_offset(*iterables, **kwargs):
         [('0', 'b'), ('1', 'c'), ('2', 'd'), ('3', 'e')]
 
     This can be used as a lightweight alternative to SciPy or pandas to analyze
-    data sets in which somes series have a lead or lag relationship.
+    data sets in which some series have a lead or lag relationship.
 
     By default, the sequence will end when the shortest iterable is exhausted.
     To continue until the longest iterable is exhausted, set *longest* to
@@ -1129,10 +1215,6 @@ def zip_offset(*iterables, **kwargs):
     sequence. Specify *fillvalue* to use some other value.
 
     """
-    offsets = kwargs['offsets']
-    longest = kwargs.get('longest', False)
-    fillvalue = kwargs.get('fillvalue', None)
-
     if len(iterables) != len(offsets):
         raise ValueError("Number of iterables and offsets didn't match")
 
@@ -1167,7 +1249,7 @@ def sort_together(iterables, key_list=(0,), reverse=False):
         [(1, 2, 3, 4), ('d', 'c', 'b', 'a')]
 
     Set a different key list to sort according to another iterable.
-    Specifying mutliple keys dictates how ties are broken::
+    Specifying multiple keys dictates how ties are broken::
 
         >>> iterables = [(3, 1, 2), (0, 1, 0), ('c', 'b', 'a')]
         >>> sort_together(iterables, key_list=(1, 2))
@@ -1184,6 +1266,54 @@ def sort_together(iterables, key_list=(0,), reverse=False):
                             reverse=reverse)))
 
 
+def unzip(iterable):
+    """The inverse of :func:`zip`, this function disaggregates the elements
+    of the zipped *iterable*.
+
+    The ``i``-th iterable contains the ``i``-th element from each element
+    of the zipped iterable. The first element is used to to determine the
+    length of the remaining elements.
+
+        >>> iterable = [('a', 1), ('b', 2), ('c', 3), ('d', 4)]
+        >>> letters, numbers = unzip(iterable)
+        >>> list(letters)
+        ['a', 'b', 'c', 'd']
+        >>> list(numbers)
+        [1, 2, 3, 4]
+
+    This is similar to using ``zip(*iterable)``, but it avoids reading
+    *iterable* into memory. Note, however, that this function uses
+    :func:`itertools.tee` and thus may require significant storage.
+
+    """
+    head, iterable = spy(iter(iterable))
+    if not head:
+        # empty iterable, e.g. zip([], [], [])
+        return ()
+    # spy returns a one-length iterable as head
+    head = head[0]
+    iterables = tee(iterable, len(head))
+
+    def itemgetter(i):
+        def getter(obj):
+            try:
+                return obj[i]
+            except IndexError:
+                # basically if we have an iterable like
+                # iter([(1, 2, 3), (4, 5), (6,)])
+                # the second unzipped iterable would fail at the third tuple
+                # since it would try to access tup[1]
+                # same with the third unzipped iterable and the second tuple
+                # to support these "improperly zipped" iterables,
+                # we create a custom itemgetter
+                # which just stops the unzipped iterables
+                # at first length mismatch
+                raise StopIteration
+        return getter
+
+    return tuple(map(itemgetter(i), it) for i, it in enumerate(iterables))
+
+
 def divide(n, iterable):
     """Divide the elements from *iterable* into *n* parts, maintaining
     order.
@@ -1228,7 +1358,7 @@ def divide(n, iterable):
     return ret
 
 
-def always_iterable(obj, base_type=(text_type, binary_type)):
+def always_iterable(obj, base_type=(str, bytes)):
     """If *obj* is iterable, return an iterator over its items::
 
         >>> obj = (1, 2, 3)
@@ -1435,7 +1565,7 @@ def count_cycle(iterable, n=None):
     return ((i, item) for i in counter for item in iterable)
 
 
-def locate(iterable, pred=bool):
+def locate(iterable, pred=bool, window_size=None):
     """Yield the index of each item in *iterable* for which *pred* returns
     ``True``.
 
@@ -1445,18 +1575,17 @@ def locate(iterable, pred=bool):
         [1, 2, 4]
 
     Set *pred* to a custom function to, e.g., find the indexes for a particular
-    item:
+    item.
 
         >>> list(locate(['a', 'b', 'c', 'b'], lambda x: x == 'b'))
         [1, 3]
 
-    Use with :func:`windowed` to find the indexes of a sub-sequence:
+    If *window_size* is given, then the *pred* function will be called with
+    that many items. This enables searching for sub-sequences:
 
-        >>> from more_itertools import windowed
         >>> iterable = [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3]
-        >>> sub = [1, 2, 3]
-        >>> pred = lambda w: w == tuple(sub)  # windowed() returns tuples
-        >>> list(locate(windowed(iterable, len(sub)), pred=pred))
+        >>> pred = lambda *args: args == (1, 2, 3)
+        >>> list(locate(iterable, pred=pred, window_size=3))
         [1, 5, 9]
 
     Use with :func:`seekable` to find indexes and then retrieve the associated
@@ -1474,7 +1603,14 @@ def locate(iterable, pred=bool):
         106
 
     """
-    return compress(count(), map(pred, iterable))
+    if window_size is None:
+        return compress(count(), map(pred, iterable))
+
+    if window_size < 1:
+        raise ValueError('window size must be at least 1')
+
+    it = windowed(iterable, window_size, fillvalue=_marker)
+    return compress(count(), starmap(pred, it))
 
 
 def lstrip(iterable, pred):
@@ -1511,13 +1647,13 @@ def rstrip(iterable, pred):
     """
     cache = []
     cache_append = cache.append
+    cache_clear = cache.clear
     for x in iterable:
         if pred(x):
             cache_append(x)
         else:
-            for y in cache:
-                yield y
-            del cache[:]
+            yield from cache
+            cache_clear()
             yield x
 
 
@@ -1606,8 +1742,7 @@ def islice_extended(iterable, *args):
                 cache.append(item)
         else:
             # When both start and stop are positive we have the normal case
-            for item in islice(it, start, stop, step):
-                yield item
+            yield from islice(it, start, stop, step)
     else:
         start = -1 if (start is None) else start
 
@@ -1652,8 +1787,7 @@ def islice_extended(iterable, *args):
 
             cache = list(islice(it, n))
 
-            for item in cache[i::step]:
-                yield item
+            yield from cache[i::step]
 
 
 def always_reversible(iterable):
@@ -1721,7 +1855,7 @@ def difference(iterable, func=sub):
 
     This is the opposite of :func:`accumulate`'s default behavior:
 
-        >>> from more_itertools import accumulate
+        >>> from itertools import accumulate
         >>> iterable = [0, 1, 2, 3, 4]
         >>> list(accumulate(iterable))
         [0, 1, 3, 6, 10]
@@ -1752,7 +1886,7 @@ def difference(iterable, func=sub):
 class SequenceView(Sequence):
     """Return a read-only view of the sequence object *target*.
 
-    :class:`SequenceView` objects are analagous to Python's built-in
+    :class:`SequenceView` objects are analogous to Python's built-in
     "dictionary view" types. They provide a dynamic view of a sequence's items,
     meaning that when the sequence updates, so does the view.
 
@@ -1793,7 +1927,7 @@ class SequenceView(Sequence):
         return '{}({})'.format(self.__class__.__name__, repr(self._target))
 
 
-class seekable(object):
+class seekable:
     """Wrap an iterator to allow for seeking backward and forward. This
     progressively caches the items in the source iterable so they can be
     re-visited.
@@ -1867,8 +2001,6 @@ class seekable(object):
         self._cache.append(item)
         return item
 
-    next = __next__
-
     def elements(self):
         return SequenceView(self._cache)
 
@@ -1879,7 +2011,7 @@ class seekable(object):
             consume(self, remainder)
 
 
-class run_length(object):
+class run_length:
     """
     :func:`run_length.encode` compresses an iterable with run-length encoding.
     It yields groups of repeated items with the count of how many times they
@@ -2032,7 +2164,7 @@ def map_reduce(iterable, keyfunc, valuefunc=None, reducefunc=None):
         [('A', 1), ('B', 2), ('C', 3)]
 
     You may want to filter the input iterable before applying the map/reduce
-    proecdure:
+    procedure:
 
         >>> all_items = range(30)
         >>> items = [x for x in all_items if 10 <= x <= 20]  # Filter
@@ -2066,3 +2198,109 @@ def map_reduce(iterable, keyfunc, valuefunc=None, reducefunc=None):
 
     ret.default_factory = None
     return ret
+
+
+def rlocate(iterable, pred=bool, window_size=None):
+    """Yield the index of each item in *iterable* for which *pred* returns
+    ``True``, starting from the right and moving left.
+
+    *pred* defaults to :func:`bool`, which will select truthy items:
+
+        >>> list(rlocate([0, 1, 1, 0, 1, 0, 0]))  # Truthy at 1, 2, and 4
+        [4, 2, 1]
+
+    Set *pred* to a custom function to, e.g., find the indexes for a particular
+    item:
+
+        >>> iterable = iter('abcb')
+        >>> pred = lambda x: x == 'b'
+        >>> list(rlocate(iterable, pred))
+        [3, 1]
+
+    If *window_size* is given, then the *pred* function will be called with
+    that many items. This enables searching for sub-sequences:
+
+        >>> iterable = [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3]
+        >>> pred = lambda *args: args == (1, 2, 3)
+        >>> list(rlocate(iterable, pred=pred, window_size=3))
+        [9, 5, 1]
+
+    Beware, this function won't return anything for infinite iterables.
+    If *iterable* is reversible, ``rlocate`` will reverse it and search from
+    the right. Otherwise, it will search from the left and return the results
+    in reverse order.
+
+    See :func:`locate` to for other example applications.
+
+    """
+    if window_size is None:
+        try:
+            len_iter = len(iterable)
+            return (
+                len_iter - i - 1 for i in locate(reversed(iterable), pred)
+            )
+        except TypeError:
+            pass
+
+    return reversed(list(locate(iterable, pred, window_size)))
+
+
+def replace(iterable, pred, substitutes, count=None, window_size=1):
+    """Yield the items from *iterable*, replacing the items for which *pred*
+    returns ``True`` with the items from the iterable *substitutes*.
+
+        >>> iterable = [1, 1, 0, 1, 1, 0, 1, 1]
+        >>> pred = lambda x: x == 0
+        >>> substitutes = (2, 3)
+        >>> list(replace(iterable, pred, substitutes))
+        [1, 1, 2, 3, 1, 1, 2, 3, 1, 1]
+
+    If *count* is given, the number of replacements will be limited:
+
+        >>> iterable = [1, 1, 0, 1, 1, 0, 1, 1, 0]
+        >>> pred = lambda x: x == 0
+        >>> substitutes = [None]
+        >>> list(replace(iterable, pred, substitutes, count=2))
+        [1, 1, None, 1, 1, None, 1, 1, 0]
+
+    Use *window_size* to control the number of items passed as arguments to
+    *pred*. This allows for locating and replacing subsequences.
+
+        >>> iterable = [0, 1, 2, 5, 0, 1, 2, 5]
+        >>> window_size = 3
+        >>> pred = lambda *args: args == (0, 1, 2)  # 3 items passed to pred
+        >>> substitutes = [3, 4] # Splice in these items
+        >>> list(replace(iterable, pred, substitutes, window_size=window_size))
+        [3, 4, 5, 3, 4, 5]
+
+    """
+    if window_size < 1:
+        raise ValueError('window_size must be at least 1')
+
+    # Save the substitutes iterable, since it's used more than once
+    substitutes = tuple(substitutes)
+
+    # Add padding such that the number of windows matches the length of the
+    # iterable
+    it = chain(iterable, [_marker] * (window_size - 1))
+    windows = windowed(it, window_size)
+
+    n = 0
+    for w in windows:
+        # If the current window matches our predicate (and we haven't hit
+        # our maximum number of replacements), splice in the substitutes
+        # and then consume the following windows that overlap with this one.
+        # For example, if the iterable is (0, 1, 2, 3, 4...)
+        # and the window size is 2, we have (0, 1), (1, 2), (2, 3)...
+        # If the predicate matches on (0, 1), we need to zap (0, 1) and (1, 2)
+        if pred(*w):
+            if (count is None) or (n < count):
+                n += 1
+                yield from substitutes
+                consume(windows, window_size - 1)
+                continue
+
+        # If there was no match (or we've reached the replacement limit),
+        # yield the first item from the window.
+        if w and (w[0] is not _marker):
+            yield w[0]
